###############################################################################
# parse GAMMA command docstrings to Python functions

# Copyright (c) 2015-2025, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
import os
import re
import subprocess as sp
from collections import Counter
from spatialist.ancillary import finder, which, dissolve

from pyroSAR.examine import ExamineGamma

import logging

log = logging.getLogger(__name__)


def parse_command(command, indent='    '):
    """
    Parse the help text of a GAMMA command to a Python function including a docstring.
    The docstring is in rst format and can thu be parsed by e.g. sphinx.
    This function is not intended to be used by itself, but rather within function :func:`parse_module`.

    Parameters
    ----------
    command: str
        the name of the gamma command
    indent: str
        the Python function indentation string; default: four spaces

    Returns
    -------
    str
        the full Python function text

    """
    # run the command without passing arguments to just catch its usage description
    command = which(command)
    if command is None:
        raise OSError('command does not exist')
    command_base = os.path.basename(command)
    proc = sp.Popen(command, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, universal_newlines=True)
    out, err = proc.communicate()
    # sometimes the description string is split between stdout and stderr
    # for the following commands stderr contains the usage description line, which is inserted into stdout
    if command_base in ['ras_pt', 'ras_data_pt', 'rasdt_cmap_pt']:
        out = out.replace(' ***\n ', ' ***\n ' + err)
    else:
        # for all other commands stderr is just appended to stdout
        out += err
    
    # raise a warning when the command has been deprecated
    # extract all lines starting and ending with three asterisks
    matches = re.findall(r'^\*{3}\s*(.*?)\s*\*{3}$', out, re.MULTILINE)
    if matches:
        # join the lines and search for a deprecation message
        cleaned = ' '.join(matches)
        pattern = (r'([\w\.]+ (?:has been|was) re(?:named to|placed(?: that [ \*\n]*|) by)'
                   r'(?:[ \*\n]*|)(?: the ISP program|) [\w\.]+)')
        match = re.search(pattern, cleaned)
        if match:
            raise DeprecationWarning(match.group())
    
    if re.search(r"Can't locate FILE/Path\.pm in @INC", out):
        raise RuntimeError('unable to parse Perl script')
    ###########################################
    # fix command-specific inconsistencies in parameter naming
    # in several commands the parameter naming in the usage description line does not match that of the docstring
    parnames_lookup = {'2PASS_INT': [('OFF_PAR', 'OFF_par')],
                       'adapt_filt': [('low_snr_thr', 'low_SNR_thr')],
                       'atm_mod2': [('rpt', 'report'),
                                    ('[mode]', '[model_atm]'),
                                    ('[model]', '[model_atm]'),
                                    ('model     atm', 'model_atm atm'),
                                    ],
                       'atm_mod_2d': [('xref', 'rref'),
                                      ('yref', 'azref')],
                       'atm_mod_2d_pt': [('[sigma_min]', '[sigma_max]')],
                       'base_calc': [('plt_flg', 'plt_flag'),
                                     ('pltflg', 'plt_flag')],
                       'base_init': [('<base>', '<baseline>')],
                       'base_plot': [('plt_flg', 'plt_flag'),
                                     ('pltflg', 'plt_flag')],
                       'cc_monitoring': [('...', '<...>')],
                       'cct_sp_pt': [('pcct_sp_pt', 'pcct_sp')],
                       'comb_interfs': [('combi_out', 'combi_int')],
                       'coord_to_sarpix': [('north/lat', 'north_lat'),
                                           ('east/lon', 'east_lon'),
                                           ('SLC_par', '<SLC_MLI_par>'),
                                           ('SLC/MLI_par', 'SLC_MLI_par')],
                       'data2geotiff': [('nodata', 'no_data')],
                       'def_mod': [('<def>', '<def_rate>'),
                                   ('def         (output)', 'def_rate    (output)')],
                       'dis2hgt': [('m/cycle', 'm_cycle')],
                       'discc': [('min_corr', 'cmin'),
                                 ('max_corr', 'cmax')],
                       'disp2ras': [('<list>', '<DISP_tab>')],
                       'dis_data': [('...', '<...>')],
                       'dispwr': [('data_type', 'dtype')],
                       'DORIS_vec': [('SLC_PAR', 'SLC_par')],
                       'gc_map_fd': [('fdtab', 'fd_tab')],
                       'gc_map_grd': [('<MLI_par>', '<GRD_par>')],
                       'geocode_back': [('<gc_map>', '<lookup_table>'),
                                        ('\n  gc_map ', '\n  lookup_table ')],
                       'GRD_to_SR': [('SLC_par', 'MLI_par')],
                       'haalpha': [('<alpha> <entropy>', '<alpha2> <entropy>'),
                                   ('alpha       (output)', 'alpha2      (output)')],
                       'histogram_ras': [('mean/stdev', 'mean_stdev')],
                       'hsi_color_scale': [('[chip]', '[chip_width]')],
                       'HUYNEN_DEC': [('T11_0', 'T11'),
                                      ('<T12> <T13> <T11>', '<T11> <T12> <T13>'),
                                      ('HUYNEN_DEC:', '***')],
                       'interf_SLC': [('  SLC2_pa  ', '  SLC2_par  ')],
                       'ionosphere_mitigation': [('<SLC1> <ID1>', '<ID1>')],
                       'landsat2dem': [('<DEM>', '<image>')],
                       'line_interp': [('input file', 'data_in'),
                                       ('output file', 'data_out')],
                       'm-alpha': [('<c2 ', '<c2> ')],
                       'm-chi': [('<c2 ', '<c2> ')],
                       'm-delta': [('<c2 ', '<c2> ')],
                       'map_section': [('n1', 'north1'),
                                       ('e1', 'east1'),
                                       ('n2', 'north2'),
                                       ('e2', 'east2'),
                                       ('[coord]', '[coords]')],
                       'mask_class': [('...', '<...>')],
                       'mcf_pt': [('<azlks>', '[azlks]'),
                                  ('<rlks>', '[rlks]')],
                       'mk_2d_im_geo': [('exponent', 'exp')],
                       'mk_adf2_2d': [('[alpha_max [', '[alpha_max] ['),
                                      ('-m MLI_dir', 'mli_dir'),
                                      ('-s scale', 'scale'),
                                      ('-e exp', 'exponent'),
                                      ('-u', 'update'),
                                      ('-D', 'dem_par')],
                       'mk_base_calc': [('<RSLC_tab>', '<SLC_tab>')],
                       'mk_cpd_all': [('dtab', 'data_tab')],
                       'mk_cpx_ref_2d': [('diff_tab', 'cpx_tab')],
                       'mk_diff_tc_2d': [('<def>', '<def_rate>'),
                                         ('def       (input)', 'def_rate  (input)')],
                       'mk_dispmap2_2d': [('RMLI_image', 'MLI'),
                                          ('RMLI_par', 'MLI_par'),
                                          ('MLI_image', 'MLI'),
                                          ('DISP_tab', 'disp_tab')],
                       'mk_dispmap_2d': [('RMLI_image', 'MLI'),
                                         ('RMLI_par', 'MLI_par'),
                                         ('MLI_image', 'MLI'),
                                         ('DISP_tab', 'disp_tab')],
                       'mk_geo_data_all': [('data_geo_dir', 'geo_dir')],
                       'mk_itab': [('<offset>', '<start>')],
                       'mk_hgt_2d': [('m/cycle', 'm_cycle')],
                       'mk_pol2rec_2d': [('data_tab', 'DIFF_tab'),
                                         ('<type> <rmli>', '<dtype>'),
                                         ('<dtype> <rmli>', '<dtype>'),
                                         ('type           input', 'dtype          input'),
                                         ('\n    Options:\n', ''),
                                         ('-s scale', 'scale'),
                                         ('-e exp', 'exponent'),
                                         ('-a min', 'min'),
                                         ('-b max', 'max'),
                                         ('-R rmax', 'rmax'),
                                         ('-m mode', 'mode'),
                                         ('-u', 'update')],
                       'mk_rasdt_all': [('RMLI_image', 'MLI'),
                                        ('MLI_image', 'MLI')],
                       'mk_rasmph_all': [('RMLI_image', 'MLI'),
                                         ('MLI_image', 'MLI')],
                       'mk_tab2': [('--linenumber', 'linenumber')],
                       'mk_unw_2d': [('unw_mask1', 'unw_mask')],
                       'mk_unw_ref_2d': [('diff_tab', 'DIFF_tab')],
                       'MLI2pt': [('MLI_TAB', 'MLI_tab'),
                                  ('pSLC_par', 'pMLI_par')],
                       'mosaic': [('<..>', '<...>'),
                                  ('DEM_parout', 'DEM_par_out')],
                       'multi_class_mapping': [('...', '<...>')],
                       'multi_def': [('<def>', '<def_rate>'),
                                     ('def             (output)', 'def_rate        (output)')],
                       'multi_look_geo': [('geo_SLC', 'SLC'),
                                          ('SLC/MLI', ('SLC_MLI'))],
                       'multi_look_MLI': [('MLI in_par', 'MLI_in_par')],
                       'offset_fit': [('interact_flag', 'interact_mode')],
                       'offset_plot_az': [('rmin', 'r_min'),
                                          ('rmax', 'r_max')],
                       'par_ASF_SLC': [('CEOS_SAR_leader', 'CEOS_leader')],
                       'par_ASAR': [('ASAR/ERS_file', 'ASAR_ERS_file')],
                       'par_EORC_JERS_SLC': [('slc', 'SLC')],
                       'par_ERSDAC_PALSAR': [('VEXCEL_SLC_par', 'ERSDAC_SLC_par')],
                       'par_ESA_JERS_SEASAT_SLC': [('[slc]', '[SLC]')],
                       'par_ICEYE_GRD': [('<GeoTIFF>', '<GeoTIFF> <XML>'),
                                         ('[mli]', '[MLI]')],
                       'par_ICEYE_SLC': [('[slc]', '[SLC]')],
                       'par_MSP': [('SLC/MLI_par', 'SLC_MLI_par')],
                       'par_SIRC': [('UTC/MET', 'UTC_MET')],
                       'par_TX_GRD': [('COSAR', 'GeoTIFF')],
                       'par_UAVSAR_SLC': [('SLC/MLC_in', 'SLC_MLC_in'),
                                          ('SLC/MLI_par', 'SLC_MLI_par'),
                                          ('SLC/MLI_out', 'SLC_MLI_out')],
                       'par_UAVSAR_geo': [('SLC/MLI_par', 'SLC_MLI_par')],
                       'phase_sim': [('sim       (', 'sim_unw   (')],
                       'product': [('wgt_flg', 'wgt_flag')],
                       'radcal_MLI': [('MLI_PAR', 'MLI_par')],
                       'radcal_PRI': [('GRD_PAR', 'GRD_par'),
                                      ('PRI_PAR', 'PRI_par')],
                       'radcal_SLC': [('SLC_PAR', 'SLC_par')],
                       'ras2jpg': [('{', '{{'),
                                   ('}', '}}')],
                       'ras_data_pt': [('pdata1', 'pdata')],
                       'ras_to_rgb': [('red channel', 'red_channel'),
                                      ('green channel', 'green_channel'),
                                      ('blue channel', 'blue_channel')],
                       'rascc_mask_thinning': [('...', '[...]')],
                       'rashgt': [('m/cycle', 'm_cycle')],
                       'rashgt_shd': [('m/cycle', 'm_cycle'),
                                      ('\n  cycle ', '\n  m_cycle ')],
                       'rasdt_cmap_pt': [('pdata1', 'pdata')],
                       'raspwr': [('hdrz', 'hdrsz')],
                       'ras_ras': [('r_lin/log', 'r_lin_log'),
                                   ('g_lin/log', 'g_lin_log'),
                                   ('b_lin/log', 'b_lin_log')],
                       'ras_ratio_dB': [('[min_cc] [max_cc] [scale] [exp]', '[min_value] [max_value] [dB_offset]')],
                       'rasSLC': [('[header]', '[hdrsz]')],
                       'ratio': [('wgt_flg', 'wgt_flag')],
                       'restore_float': [('input file', 'data_in'),
                                         ('output file', 'data_out'),
                                         ('interpolation_limit', 'interp_limit')],
                       'S1_coreg_TOPS_no_refinement': [('RLK', 'rlks'),
                                                       ('AZLK', 'azlks')],
                       'S1_OPOD_vec': [('SLC_PAR', 'SLC_par')],
                       'single_class_mapping': [('>...', '> <...>')],
                       'ScanSAR_burst_cc_ad': [('bx', 'box_min'),
                                               ('by', 'box_max')],
                       'ScanSAR_burst_to_mosaic': [('DATA_tab_ref', 'data_tab_ref'),
                                                   ('[mflg] [dtype]', '[mflg]')],
                       'ScanSAR_full_aperture_SLC': [('SLCR_dir', 'SLC2_dir')],
                       'scale_base': [('SLC-1_par-2', 'SLC1_par-2')],
                       'sigma2gamma': [('<gamma>', '<gamma0>'),
                                       ('gamma  (output)', 'gamma0  (output)'),
                                       ('pwr1', 'sigma0')],
                       'SLC_interp_lt': [('SLC-2', 'SLC2'),
                                         ('blksz', 'blk_size')],
                       'SLC_intf': [('SLC1s_par', 'SLC-1s_par'),
                                    ('SLC2Rs_par', 'SLC-2Rs_par')],
                       'SLC_intf_geo2': [('cc        (', 'CC        (')],
                       'SLC_interp_map': [('coffs2_sm', 'coffs_sm')],
                       'SLC_mosaic_S1_TOPS': [('wflg', 'bflg')],
                       'srtm_mosaic': [('<lon>', '<lon2>')],
                       'SSI_INT_S1': [('<SLC2> <par2>', '<SLC_tab2>')],
                       'texture': [('weights_flag', 'wgt_flag')],
                       'ts_rate': [('sim_flg', 'sim_flag')],
                       'TX_SLC_preproc': [('TX_list', 'TSX_list')],
                       'uchar2float': [('infile', 'data_in'),
                                       ('outfile', 'data_out')],
                       'validate': [('ras1', 'ras_map'),
                                    ('rasf_map', 'ras_map'),
                                    ('ras2', 'ras_inv'),
                                    ('rasf_inventory', 'ras_inv'),
                                    ('class1[1]', 'class1_1'),
                                    ('class1[2]', 'class1_2'),
                                    ('class1[n]', 'class1_n'),
                                    ('class2[1]', 'class2_1'),
                                    ('class2[2]', 'class2_2'),
                                    ('class2[n]', 'class2_n')]}
    if command_base in parnames_lookup.keys():
        for replacement in parnames_lookup[command_base]:
            out = out.replace(*replacement)
    ###########################################
    # filter header (general command description) and usage description string
    header = '\n'.join([x.strip('* ') for x in re.findall('[*]{3}.*(?:[*]{3}|)', out)])
    header = '| ' + header.replace('\n', '\n| ')
    usage = re.search('usage:.*(?=\n)', out).group()
    
    # filter required and optional arguments from usage description text
    arg_req_raw = [re.sub(r'[^\w.-]*', '', x) for x in re.findall('[^<]*<([^>]*)>', usage)]
    arg_opt_raw = [re.sub(r'[^\w.-]*', '', x) for x in re.findall(r'[^[]*\[([^]]*)]', usage)]
    
    ###########################################
    # add parameters missing in the usage argument lists
    
    appends = {'mk_adf2_2d': ['cc_min', 'cc_max', 'mli_dir', 'scale', 'exponent', 'update', 'dem_par'],
               'mk_pol2rec_2d': ['scale', 'exponent', 'min', 'max', 'rmax', 'mode', 'update'],
               'SLC_interp_S1_TOPS': ['mode', 'order'],
               'SLC_interp_map': ['mode', 'order']}
    
    if command_base in appends.keys():
        for var in appends[command_base]:
            if var not in arg_opt_raw:
                arg_opt_raw.append(var)
    ###########################################
    # define parameter replacements; this is intended for parameters which are to be aggregated into a list parameter
    replacements = {'cc_monitoring': [(['nfiles', 'f1', 'f2', '...'],
                                       ['files'],
                                       ['a list of input data files (float)'])],
                    'dis_data': [(['nstack', 'pdata1', '...'],
                                  ['pdata'],
                                  ['a list of point data stack files'])],
                    'lin_comb': [(['nfiles', 'f1', 'f2', '...'],
                                  ['files'],
                                  ['a list of input data files (float)']),
                                 (['factor1', 'factor2', '...'],
                                  ['factors'],
                                  ['a list of factors to multiply the input files with'])],
                    'lin_comb_cpx': [(['nfiles', 'f1', 'f2', '...'],
                                      ['files'],
                                      ['a list of input data files (float)']),
                                     (['factor1_r', 'factor2_r', '...'],
                                      ['factors_r'],
                                      ['a list of real part factors to multiply the input files with']),
                                     (['factor1_i', 'factor2_i'],
                                      ['factors_i'],
                                      ['a list of imaginary part factors to multiply the input files with'])],
                    'mask_class': [(['n_class', 'class_1', '...', 'class_n'],
                                    ['class_values'],
                                    ['a list of class map values'])],
                    'mosaic': [(['nfiles', 'data_in1', 'DEM_par1', 'data_in2', 'DEM_par2', '...', '...'],
                                ['data_in_list', 'DEM_par_list'],
                                ['a list of input data files',
                                 'a list of DEM/MAP parameter files for each data file'])],
                    'multi_class_mapping': [(['nfiles', 'f1', 'f2', '...', 'fn'],
                                             ['files'],
                                             ['a list of input data files (float)'])],
                    'rascc_mask_thinning': [(['thresh_1', '...', 'thresh_nmax'],
                                             ['thresholds'],
                                             ['a list of thresholds sorted from smallest to '
                                              'largest scale sampling reduction'])],
                    'single_class_mapping': [(['nfiles', 'f1', '...', 'fn'],
                                              ['files'],
                                              ['a list of point data stack files']),
                                             (['lt1', 'ltn'],
                                              ['thres_lower'],
                                              ['a list of lower thresholds for the files']),
                                             (['ut1', 'utn'],
                                              ['thres_upper'],
                                              ['a list of upper thresholds for the files'])],
                    'validate': [(['nclass1', 'class1_1', 'class1_2', '...', 'class1_n'],
                                  ['classes_map'],
                                  ['a list of class values for the map data file (max. 16), 0 for all']),
                                 (['nclass2', 'class2_1', 'class2_2', '...', 'class2_n'],
                                  ['classes_inv'],
                                  ['a list of class values for the inventory data file (max. 16), 0 for all'])]}
    
    if '..' in usage and command_base not in replacements.keys():
        raise RuntimeError('the command contains multi-args which were not properly parsed')
    
    def replace(inlist, replacement):
        outlist = list(inlist)
        for old, new, description in replacement:
            if old[0] not in outlist:
                return outlist
            outlist[outlist.index(old[0])] = new
            for i in range(1, len(old)):
                if old[i] in outlist:
                    outlist.remove(old[i])
        return dissolve(outlist)
    
    arg_req = list(arg_req_raw)
    arg_opt = list(arg_opt_raw)
    
    if command_base in replacements.keys():
        arg_req = replace(arg_req, replacements[command_base])
        arg_opt = replace(arg_opt, replacements[command_base])
    ###########################################
    # check if there are any double parameters
    
    double = [k for k, v in Counter(arg_req + arg_opt).items() if v > 1]
    if len(double) > 0:
        raise RuntimeError('double parameter{0}: {1}'.format('s' if len(double) > 1 else '', ', '.join(double)))
    ###########################################
    # add a parameter inlist for commands which take interactive input via stdin
    # the list of commands, which are interactive is hard to assess and thus likely a source of future errors
    
    inlist = ['create_dem_par', 'par_ESA_ERS']
    
    if command_base in inlist:
        arg_req.append('inlist')
    
    ######################################################################################
    # create the function argument string for the Python function
    
    # optional arguments are parametrized with '-' as default value, e.g., arg_opt='-'
    argstr_function = ', '.join(arg_req + [x + "='-'" for x in arg_opt])
    # a '-' in the parameter name is replaced with '_'
    argstr_function = re.sub(r'([^\'])-([^\'])', r'\1_\2', argstr_function)
    # replace unsupported 'def' parameter name with 'drm'
    argstr_function = argstr_function.replace(', def=', ', drm=')
    
    # some commands have different defaults than '-'
    replacements_defaults = {'S1_import_SLC_from_zipfiles': {'OPOD_dir': '.'}}
    
    if command_base in replacements_defaults.keys():
        for key, value in replacements_defaults[command_base].items():
            old = f"{key}='-'"
            if isinstance(value, str):
                new = f"{key}='{value}'"
            else:
                new = f"{key}={value}"
            argstr_function = argstr_function.replace(old, new)
    
    # create the function definition string
    fun_def = 'def {name}({args_fun}, logpath=None, outdir=None, shellscript=None):' \
        .format(name=command_base.replace('-', '_'),
                args_fun=argstr_function)
    
    if command_base == '2PASS_INT':
        fun_def = fun_def.replace(command_base, 'TWO_PASS_INT')
    ######################################################################################
    # special handling of flag args
    flag_args = {'mk_adf2_2d': [('mli_dir', '-m', None),
                                ('scale', '-s', None),
                                ('exponent', '-e', None),
                                ('update', '-u', False),
                                ('dem_par', '-D', None)],
                 'mk_pol2rec_2d': [('scale', '-s', None),
                                   ('exp', '-e', None),
                                   ('min', '-a', None),
                                   ('max', '-b', None),
                                   ('rmax', '-R', None),
                                   ('mode', '-m', None),
                                   ('update', '-u', False)],
                 'mk_tab2': [('linenumber', '--linenumber', False)]}
    
    # replace arg default like arg='-' with arg=None or arg=False
    if command_base in flag_args:
        for arg in flag_args[command_base]:
            fun_def = re.sub(f'{arg[0]}=\'-\'', f'{arg[0]}={arg[2]}', fun_def)
    ######################################################################################
    # create the process call argument string
    
    # a '-' in the parameter name is replaced with '_'
    # e.g. 'arg1, arg2, arg3'
    # if a parameter is named 'def' (not allowed in Python) it is renamed to 'drm'
    
    # inlist is not a proc arg but a parameter passed to function process
    proc_args = arg_req + arg_opt
    if command_base in inlist:
        proc_args.remove('inlist')
    proc_args_tmp = list(proc_args)
    # insert the length of a list argument as a proc arg
    if command_base in replacements.keys() and command_base != 'rascc_mask_thinning':
        key = replacements[command_base][0][1]
        if isinstance(key, list):
            key = key[0]
        proc_args_tmp.insert(proc_args_tmp.index(key), f'len({key})')
    
    if command_base == 'validate':
        index = proc_args_tmp.index('classes_inv')
        proc_args_tmp.insert(index, 'len(classes_inv)')
    
    argstr_process = ', '.join(proc_args_tmp) \
        .replace('-', '_') \
        .replace(', def,', ', drm,')
    
    # create the process argument list string
    cmd_str = "cmd = ['{command}', {args_cmd}]".format(command=command, args_cmd=argstr_process)
    
    # special handling of optional flag args
    # the args are removed from the cmd list and flags (plus values) added if not None or True
    # e.g. '-u' if update=True or '-m /path' if mli_dir='/path'
    if command_base in flag_args:
        args = []
        for arg in flag_args[command_base]:
            cmd_str = cmd_str.replace(', {}'.format(arg[0]), '')
            args.append(arg[0])
            cmd_str += "\nif {a} is not {d}:\n{i}cmd.append('{k}')" \
                .format(i=indent, d=arg[2], k=arg[1], a=arg[0])
            if arg[2] is None:
                cmd_str += '\n{i}cmd.append({a})'.format(i=indent, a=arg[0])
    
    # create the process call string
    proc_str = "process(cmd, logpath=logpath, outdir=outdir{inlist}, shellscript=shellscript)" \
        .format(inlist=', inlist=inlist' if command_base in inlist else '')
    fun_proc = '{0}\n{1}'.format(cmd_str, proc_str)
    
    if command_base == 'lin_comb_cpx':
        fun_proc = fun_proc.replace('factors_r, factors_i', 'zip(factors_r, factors_i)')
    elif command_base == 'mosaic':
        fun_proc = fun_proc.replace('data_in_list, DEM_par_list', 'zip(data_in_list, DEM_par_list)')
    elif command_base == 'single_class_mapping':
        fun_proc = fun_proc.replace('files, thres_lower, thres_upper', 'zip(files, thres_lower, thres_upper)')
    
    ######################################################################################
    # create the function docstring
    
    # find the start of the docstring and filter the result
    doc_start = 'input parameters:[ ]*\n' if re.search('input parameters', out) else 'usage:.*(?=\n)'
    doc = '\n' + out[re.search(doc_start, out).end():]
    
    # define a pattern containing individual parameter documentations
    pattern = r'\n[ ]*[<\[]*(?P<par>{0})[>\]]*[\t ]+(?P<doc>.*)'.format(
        '|'.join(arg_req_raw + arg_opt_raw).replace('.', r'\.'))
    
    # identify the start indices of all pattern matches
    starts = [m.start(0) for m in re.finditer(pattern, doc)] + [len(out)]
    
    # filter out all individual (parameter, description) docstring tuples
    doc_items = []
    j = 0
    done = []
    for i in range(0, len(starts) - 1):
        doc_raw = doc[starts[i]:starts[i + 1]]
        doc_list = list(re.search(pattern, doc_raw, flags=re.DOTALL).groups())
        
        if doc_list[0] not in proc_args:
            if command_base in replacements.keys():
                repl = replacements[command_base][0]
                for k, item in enumerate(repl[1]):
                    if item not in done:
                        doc_items.append([item, repl[2][k]])
                        done.append(item)
                        j += 1
            continue
        
        if doc_list[0] in done:
            doc_items[-1][1] += doc_raw
            continue
        
        while doc_list[0] != proc_args[j]:
            doc_list_sub = [proc_args[j], 'not documented']
            doc_items.append(doc_list_sub)
            j += 1
        
        doc_items.append(doc_list)
        done.append(doc_items[-1][0])
        j += 1
    
    for k in range(j, len(proc_args)):
        doc_items.append([proc_args[k], 'not documented'])
    
    # add a parameter inlist to the docstring tuples
    if command_base in inlist:
        pos = [x[0] for x in doc_items].index(arg_opt[0])
        doc_items.insert(pos, ('inlist', 'a list of arguments to be passed to stdin'))
    
    # remove the replaced parameters from the argument lists
    doc_items = [x for x in doc_items if x[0] in arg_req + arg_opt]
    
    # replace parameter names which are not possible in Python syntax, i.e. containing '-' or named 'def'
    for i, item in enumerate(doc_items):
        par = item[0].replace('-', '_').replace(', def,', ', drm,')
        description = item[1]
        doc_items[i] = (par, description)
    
    if command_base in ['par_CS_geo', 'par_KS_geo']:
        doc_items.append(('MLI_par', '(output) ISP SLC/MLI parameter file (example: yyyymmdd.mli.par)'))
        doc_items.append(('DEM_par', '(output) DIFF/GEO DEM parameter file (example: yyyymmdd.dem_par)'))
        doc_items.append(('GEO', '(output) Geocoded image data file (example: yyyymmdd.geo)'))
    
    # check if all parameters are documented:
    proc_args = [x.replace('-', '_').replace(', def,', ', drm,') for x in arg_req + arg_opt]
    mismatch = [x for x in proc_args if x not in [y[0] for y in doc_items]]
    if len(mismatch) > 0:
        raise RuntimeError('parameters missing in docsring: {}'.format(', '.join(mismatch)))
    ###########################################
    # format the docstring parameter descriptions
    
    docstring_elements = ['Parameters\n----------']
    
    # do some extra formatting
    for i, item in enumerate(doc_items):
        par, description = item
        description = re.split(r'\n+\s*', description.strip('\n'))
        
        # escape * characters (which are treated as special characters for bullet lists by sphinx)
        description = [x.replace('*', r'\\*') for x in description]
        
        # convert all lines starting with an integer number or 'NOTE' to bullet list items
        latest = None
        for i in range(len(description)):
            item = description[i]
            if re.search('^(?:(?:-|)[-0-9]+|NOTE):', item):
                latest = i
                # prepend '* ' and replace missing spaces after a colon: 'x:x' -> 'x: x'
                description[i] = '* ' + re.sub(r'((?:-|)[-0-9]+:)(\w+)', r'\1 \2', item)
        
        # format documentation lines coming after the last bullet list item
        # sphinx expects lines after the last bullet item to be indented by two spaces if
        # they belong to the bullet item or otherwise a blank line to mark the end of the bullet list
        if latest:
            # case if there are still lines coming after the last bullet item,
            # prepend an extra two spaces to these lines so that they are properly
            # aligned with the text of the bullet item
            if latest + 2 <= len(description):
                i = 1
                while latest + i + 1 <= len(description):
                    description[latest + i] = '  ' + description[latest + i]
                    i += 1
            # if not, then insert an extra blank line
            else:
                description[-1] = description[-1] + '\n'
        
        # parse the final documentation string for the current parameter
        description = '\n{0}{0}'.join(description).format(indent)
        doc = '{0}:\n{1}{2}'.format(par, indent, description)
        docstring_elements.append(doc)
    ###########################################
    # add docsrings of general parameters and combine the result
    
    # create docstring for parameter logpath
    doc = 'logpath: str or None\n{0}a directory to write command logfiles to'.format(indent)
    docstring_elements.append(doc)
    
    # create docstring for parameter outdir
    doc = 'outdir: str or None\n{0}the directory to execute the command in'.format(indent)
    docstring_elements.append(doc)
    
    # create docstring for parameter shellscript
    doc = 'shellscript: str or None\n{0}a file to write the Gamma commands to in shell format'.format(indent)
    docstring_elements.append(doc)
    
    # combine the complete docstring
    fun_doc = '\n{header}\n\n{doc}\n' \
        .format(header=header,
                doc='\n'.join(docstring_elements))
    ######################################################################################
    
    # combine the elements to a complete Python function string
    fun = '''{defn}\n"""{doc}"""\n{proc}'''.format(defn=fun_def, doc=fun_doc, proc=fun_proc)
    
    # indent all lines and add an extra empty line at the end
    fun = fun.replace('\n', '\n{}'.format(indent)) + '\n'
    
    return fun


def parse_module(bindir, outfile):
    """
    parse all Gamma commands of a module to functions and save them to a Python script.

    Parameters
    ----------
    bindir: str
        the `bin` directory of a module containing the commands
    outfile: str
        the name of the Python file to write

    Returns
    -------
    
    Examples
    --------
    >>> import os
    >>> from pyroSAR.gamma.parser import parse_module
    >>> outname = os.path.join(os.environ['HOME'], 'isp.py')
    >>> parse_module('/cluster/GAMMA_SOFTWARE-20161207/ISP/bin', outname)
    """
    
    if not os.path.isdir(bindir):
        raise OSError('directory does not exist: {}'.format(bindir))
    
    excludes = ['coord_trans',  # doesn't take any parameters and is interactive
                'RSAT2_SLC_preproc',  # takes option flags
                'mk_ASF_CEOS_list',  # "cannot create: Directory nonexistent"
                '2PASS_UNW',  # parameter name inconsistencies
                'mk_diff_2d',  # takes option flags
                'gamma_doc'  # opens the Gamma documentation
                ]
    failed = []
    outstring = ''
    for cmd in sorted(finder(bindir, [r'^\w+$'], regex=True), key=lambda s: s.lower()):
        basename = os.path.basename(cmd)
        if basename not in excludes:
            try:
                fun = parse_command(cmd)
            except RuntimeError as e:
                failed.append('{0}: {1}'.format(basename, str(e)))
                continue
            except DeprecationWarning:
                continue
            except:
                failed.append('{0}: {1}'.format(basename, 'error yet to be assessed'))
                continue
            outstring += fun + '\n\n'
    if len(outstring) > 0:
        if not os.path.isfile(outfile):
            with open(outfile, 'w') as out:
                out.write('from pyroSAR.gamma.auxil import process\n\n\n')
        with open(outfile, 'a') as out:
            out.write(outstring)
    if len(failed) > 0:
        info = 'the following functions could not be parsed:\n{0}\n({1} total)'
        log.info(info.format('\n'.join(failed), len(failed)))


def autoparse():
    """
    automatic parsing of GAMMA commands.
    This function will detect the GAMMA installation via environment variable `GAMMA_HOME`, detect all available
    modules (e.g. ISP, DIFF) and parse all the module's commands via function :func:`parse_module`.
    A new Python module will be created called `gammaparse`, which is stored under `$HOME/.pyrosar`.
    Upon importing the `pyroSAR.gamma` submodule, this function is run automatically and module `gammaparse`
    is imported as `api`.
    
    Returns
    -------

    Examples
    --------
    >>> from pyroSAR.gamma.api import diff
    >>> print('create_dem_par' in dir(diff))
    True
    """
    home = ExamineGamma().home
    target = os.path.join(os.path.expanduser('~'), '.pyrosar', 'gammaparse')
    if not os.path.isdir(target):
        os.makedirs(target)
    for module in finder(home, ['[A-Z]*'], foldermode=2):
        outfile = os.path.join(target, os.path.basename(module).lower() + '.py')
        if not os.path.isfile(outfile):
            log.info('parsing module {} to {}'.format(os.path.basename(module), outfile))
            for submodule in ['bin', 'scripts']:
                log.info(submodule)
                try:
                    parse_module(os.path.join(module, submodule), outfile)
                except OSError:
                    log.info('..does not exist')
    modules = [re.sub(r'\.py', '', os.path.basename(x)) for x in finder(target, [r'[a-z]+\.py$'], regex=True)]
    if len(modules) > 0:
        with open(os.path.join(target, '__init__.py'), 'w') as init:
            init.write('from . import {}'.format(', '.join(modules)))



###############################################################################
# tools for processing ERS satellite data

# Copyright (c) 2014-2019, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
################################################################################
import os
import math
from spatialist import sqlite_setup
from spatialist.ancillary import HiddenPrints
from datetime import datetime, timedelta

import logging
log = logging.getLogger(__name__)


def passdb_create(ers1passes, ers2passes, dbname):
    """
    create a sqlite database from ERS pass tables
    downloaded from http://www.deos.tudelft.nl/ers/phases/starttimes.html.
    There you can also find additional information on the file structure and background.
    The fields `phase`, `cycle`, `pass`, `starttime` and `endtime` are read from the table.
    The latter two are converted to format YYYY-MM-DD HH:MM:SS.SSS.
    The fields `cycle` and `pass` are converted to integer.
    All five fields plus the name of the sensor (`ERS1` or `ERS2`) are then stored to the database.

    Parameters
    ----------
    ers1passes: str
        the name of the ERS-1 pass table
    ers2passes: str
        the name of the ERS-2 pass table
    dbname: str
        the name of the database to write the results to

    Returns
    -------

    """
    columns = {'satellite': 'TEXT',
               'phase': 'TEXT',
               'cycleNumber': 'INTEGER',
               'passNumber': 'INTEGER',
               'starttime': 'TEXT',
               'endtime': 'TEXT'}
    
    con = sqlite_setup(driver=dbname)
    
    create_string = '''CREATE TABLE if not exists data ({})'''.format(
        ', '.join([' '.join(x) for x in columns.items()]))
    cursor = con.cursor()
    cursor.execute(create_string)
    
    def time_convert(timestring):
        dt = datetime(1985, 1, 1) + timedelta(seconds=float(timestring))
        return dt.strftime('%Y-%m-%d %H:%M:%S.%f')
    
    insert_string = '''INSERT INTO data({0}) VALUES({1})''' \
        .format(', '.join(columns.keys()),
                ', '.join(['?'] * len(columns.keys())))
    
    for satellite, filename in [('ERS1', ers1passes), ('ERS2', ers2passes)]:
        with open(filename, 'r') as table:
            for line in table:
                phase, cycle, passNumber, starttime, endtime = line.split()[0:5]
                insertion = [satellite, phase,
                             int(cycle), int(passNumber),
                             time_convert(starttime), time_convert(endtime)]
                if satellite == 'ERS1':
                    log.info(tuple(insertion))
                cursor.execute(insert_string, tuple(insertion))
    con.commit()
    con.close()


def passdb_query(satellite, acqtime, dbname=None):
    """
    query the orbit information for an ERS acquisition

    Parameters
    ----------
    satellite: {'ERS1', 'ERS2'}
        the name of the satellite
    acqtime: datetime.datetime
        the acquisition of the satellite image
    dbname: str, None
        the name of the database as created by :func:`passdb_create`. If None, the default database delivered with
        pyroSAR is used

    Returns
    -------

    """
    if satellite == 'ERS1':
        # the last timestamp for which specific ERS-1 orbit information is present,
        # afterwards that of ERS-2 is used
        last = datetime.strptime('1996-06-02 21:59:26.618659', '%Y-%m-%d %H:%M:%S.%f')
        sat = 'ERS2' if acqtime > last else 'ERS1'
    elif satellite == 'ERS2':
        sat = 'ERS2'
    else:
        raise ValueError("satellite must be either 'ERS1' or 'ERS2', was '{}'".format(satellite))
    
    if dbname is None:
        dbname = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'erspasses.db')
    with HiddenPrints():
        con = sqlite_setup(driver=dbname)
    
    cursor = con.cursor()
    acqtime_str = acqtime.strftime('%Y-%m-%d %H:%M:%S.%f')
    query = '''SELECT * FROM data WHERE satellite = ? AND starttime <= ? AND endtime >= ?'''
    cursor.execute(query, (sat, acqtime_str, acqtime_str))
    
    fetch = cursor.fetchall()
    if len(fetch) == 0:
        cursor.execute(query, ('ERS2', acqtime_str, acqtime_str))
        fetch = cursor.fetchall()
    
    result = dict(zip(['satellite', 'phase', 'cycleNumber', 'passNumber'], fetch[0][0:4]))
    result['satellite'] = satellite
    result['orbitNumber_rel'] = int(math.ceil(result['passNumber'] / 2.))
    return result



###############################################################################
# import wrapper for the pyroSAR GAMMA API

# Copyright (c) 2018-2019, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
import os
import sys
import warnings

from .parser import autoparse

try:
    autoparse()
    
    sys.path.insert(0, os.path.join(os.path.expanduser('~'), '.pyrosar'))
    
    try:
        from gammaparse import *
    except ImportError:
        warnings.warn('found a GAMMA installation directory, but module parsing failed')

except RuntimeError:
    warnings.warn('could not find GAMMA installation directory; please set the GAMMA_HOME environment variable')



###############################################################################
# Reading and Organizing system for SAR images
# Copyright (c) 2016-2023, the pyroSAR Developers.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
###############################################################################
"""
This file contains regular expressions to identify SAR products.
The pattern 'pyrosar' identifies products in pyroSAR's unified naming scheme.
The names of all other expressions correspond to the classes found in pyroSAR.drivers.
"""
pyrosar = r'(?:.*[/\\]|)' \
          r'(?P<outname_base>' \
          r'(?P<sensor>[A-Z0-9]{1,4})_+' \
          r'(?P<acquisition_mode>[A-Z0-9]{1,4})_+' \
          r'(?P<orbit>[AD])_' \
          r'(?P<start>[0-9T]{15})' \
          r'(?:_(?P<extensions>\w*?)|)' \
          r')_*' \
          r'(?:(?P<polarization>[HV]{2})_' \
          r'(?P<proc_steps>[\w-]*)|)' \
          r'(?P<filetype>(?:.tif|.nc|))$'

ceos_ers = r'(?P<product_id>(?:SAR|ASA)_(?:IM(?:S|P|G|M|_)|AP(?:S|P|G|M|_)|WV(?:I|S|W|_)|WS(?:M|S|_))_[012B][CP])' \
           r'(?P<processing_stage_flag>[A-Z])' \
           r'(?P<originator_ID>[A-Z\-]{3})' \
           r'(?P<start_day>[0-9]{8})_' \
           r'(?P<start_time>[0-9]{6})_' \
           r'(?P<duration>[0-9]{8})' \
           r'(?P<phase>[0-9A-Z]{1})' \
           r'(?P<cycle>[0-9]{3})_' \
           r'(?P<relative_orbit>[0-9]{5})_' \
           r'(?P<absolute_orbit>[0-9]{5})_' \
           r'(?P<counter>[0-9]{4,})\.' \
           r'(?P<satellite_ID>[EN][12])' \
           r'(?P<extension>(?:\.zip|\.tar\.gz|\.PS|))$'

ceos_psr1 = r'^LED-ALPSR' \
            r'(?P<sub>P|S)' \
            r'(?P<orbit>[0-9]{5})' \
            r'(?P<frame>[0-9]{4})-' \
            r'(?P<mode>[HWDPC])' \
            r'(?P<level>1\.[015])' \
            r'(?P<proc>G|_)' \
            r'(?P<proj>[UPML_])' \
            r'(?P<orbit_dir>A|D)$'

ceos_psr2 = r'^LED-ALOS2' \
            r'(?P<orbit>[0-9]{5})' \
            r'(?P<frame>[0-9]{4})-' \
            r'(?P<date>[0-9]{6})-' \
            r'(?P<mode>SBS|UBS|UBD|HBS|HBD|HBQ|FBS|FBD|FBQ|WBS|WBD|WWS|WWD|VBS|VBD)' \
            r'(?P<look_dir>L|R)' \
            r'(?P<level>1\.0|1\.1|1\.5|2\.1|3\.1)' \
            r'(?P<proc>[GR_])' \
            r'(?P<proj>[UPML_])' \
            r'(?P<orbit_dir>A|D)$'

eorc_psr = r'^PSR2-' \
           r'(?P<prodlevel>SLTR)_' \
           r'(?P<pathnr>RSP[0-9]{3})_' \
           r'(?P<date>[0-9]{8})' \
           r'(?P<mode>FBD|WBD)' \
           r'(?P<beam>[0-9]{2})' \
           r'(?P<orbit_dir>A|D)' \
           r'(?P<look_dir>L|R)_' \
           r'(?P<replay_id1>[0-9A-Z]{16})-' \
           r'(?P<replay_id2>[0-9A-Z]{5})_' \
           r'(?P<internal>[0-9]{3})_' \
           r'HDR$'

esa = r'(?P<product_id>(?:SAR|ASA)_(?:IM(?:S|P|G|M|_)|AP(?:S|P|G|M|_)|WV(?:I|S|W|_)|WS(?:M|S|_))_[012B][CP])' \
      r'(?P<processing_stage_flag>[A-Z])' \
      r'(?P<originator_ID>[A-Z\-]{3})' \
      r'(?P<start_day>[0-9]{8})_' \
      r'(?P<start_time>[0-9]{6})_' \
      r'(?P<duration>[0-9]{8})' \
      r'(?P<phase>[0-9A-Z]{1})' \
      r'(?P<cycle>[0-9]{3})_' \
      r'(?P<relative_orbit>[0-9]{5})_' \
      r'(?P<absolute_orbit>[0-9]{5})_' \
      r'(?P<counter>[0-9]{4,})\.' \
      r'(?P<satellite_ID>[EN][12])'

safe = r'^(?P<sensor>S1[ABCD])_' \
       r'(?P<beam>S1|S2|S3|S4|S5|S6|IW|EW|WV|EN|N1|N2|N3|N4|N5|N6|IM)_' \
       r'(?P<product>SLC|GRD|OCN)' \
       r'(?P<resolution>F|H|M|_)_' \
       r'(?P<processingLevel>1|2)' \
       r'(?P<category>S|A)' \
       r'(?P<pols>SH|SV|DH|DV|VV|HH|HV|VH)_' \
       r'(?P<start>[0-9]{8}T[0-9]{6})_' \
       r'(?P<stop>[0-9]{8}T[0-9]{6})_' \
       r'(?P<orbitNumber>[0-9]{6})_' \
       r'(?P<dataTakeID>[0-9A-F]{6})_' \
       r'(?P<productIdentifier>[0-9A-F]{4})' \
       r'\.SAFE$'

tsx = r'^(?P<sat>T[DS]X1)_SAR__' \
      r'(?P<prod>SSC|MGD|GEC|EEC)_' \
      r'(?P<var>____|SE__|RE__|MON1|MON2|BTX1|BRX2)_' \
      r'(?P<mode>SM|SL|HS|HS300|ST|SC)_' \
      r'(?P<pols>[SDTQ])_' \
      r'(?:SRA|DRA)_' \
      r'(?P<start>[0-9]{8}T[0-9]{6})_' \
      r'(?P<stop>[0-9]{8}T[0-9]{6})(?:\.xml|)$'

tdm = r'^(?P<sat>T[D]M1)_SAR__' \
      r'(?P<prod>COS)_' \
      r'(?P<var>____|MONO|BIST|ALT1|ALT2)_' \
      r'(?P<mode>SM|SL|HS)_' \
      r'(?P<pols>[SDQ])_' \
      r'(?:SRA|DRA)_' \
      r'(?P<start>[0-9]{8}T[0-9]{6})_' \
      r'(?P<stop>[0-9]{8}T[0-9]{6})(?:\.xml|)$'



"""
Visvalingam-Whyatt method of poly-line vertex reduction
Visvalingam, M and Whyatt J D (1993)
"Line Generalisation by Repeated Elimination of Points", Cartographic J., 30 (1), 46 - 51
Described here:
https://web.archive.org/web/20100428020453/http://www2.dcs.hull.ac.uk/CISRG/publications/DPs/DP10/DP10.html
=========================================
The MIT License (MIT)
Copyright (c) 2014 Elliot Hallmark
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
================================

code was obtained from https://github.com/Permafacture/Py-Visvalingam-Whyatt/blob/master/polysimplify.py
minor edits for Python3 compatibility by John Truckenbrodt 2019
"""

from numpy import array, argmin
import numpy as np

import logging
log = logging.getLogger(__name__)


def triangle_area(p1, p2, p3):
    """
    calculates the area of a triangle given its vertices
    """
    return abs(p1[0] * (p2[1] - p3[1]) + p2[0] * (p3[1] - p1[1]) + p3[0] * (p1[1] - p2[1])) / 2.


def triangle_areas_from_array(arr):
    """
    take an (N,2) array of points and return an (N,1)
    array of the areas of those triangles, where the first
    and last areas are np.inf
    see triangle_area for algorithm
    """

    result = np.empty((len(arr),), arr.dtype)
    result[0] = np.inf
    result[-1] = np.inf

    p1 = arr[:-2]
    p2 = arr[1:-1]
    p3 = arr[2:]

    # an accumulators to avoid unnecessary intermediate arrays
    accr = result[1:-1]  # Accumulate directly into result
    acc1 = np.empty_like(accr)

    np.subtract(p2[:, 1], p3[:, 1], out=accr)
    np.multiply(p1[:, 0], accr, out=accr)
    np.subtract(p3[:, 1], p1[:, 1], out=acc1)
    np.multiply(p2[:, 0], acc1, out=acc1)
    np.add(acc1, accr, out=accr)
    np.subtract(p1[:, 1], p2[:, 1], out=acc1)
    np.multiply(p3[:, 0], acc1, out=acc1)
    np.add(acc1, accr, out=accr)
    np.abs(accr, out=accr)
    accr /= 2.
    # Notice: accr was writing into result, so the answer is in there
    return result


# the final value in thresholds is np.inf, which will never be
# the min value.  So, I am safe in "deleting" an index by
# just shifting the array over on top of it
def remove(s, i):
    """
    Quick trick to remove an item from a numpy array without
    creating a new object.  Rather than the array shape changing,
    the final value just gets repeated to fill the space.
    ~3.5x faster than numpy.delete
    """
    s[i:-1] = s[i + 1:]


class VWSimplifier(object):
    def __init__(self, pts):
        """
        Initialize with points. takes some time to build
        the thresholds but then all threshold filtering later
        is ultra fast
        """
        self.pts = np.array(pts)
        self.thresholds = self.build_thresholds()
        self.ordered_thresholds = sorted(self.thresholds, reverse=True)

    def build_thresholds(self):
        """
        compute the area value of each vertex, which one would
        use to mask an array of points for any threshold value.
        returns a numpy.array (length of pts)  of the areas.
        """
        pts = self.pts
        nmax = len(pts)
        real_areas = triangle_areas_from_array(pts)
        real_indices = list(range(nmax))

        # destructable copies
        # ARG! areas=real_areas[:] doesn't make a copy!
        areas = np.copy(real_areas)
        i = real_indices[:]

        # pick first point and set up for loop
        min_vert = int(argmin(areas))
        this_area = areas[min_vert]
        #  areas and i are modified for each point finished
        remove(areas, min_vert)  # faster
        # areas = np.delete(areas,min_vert) #slower
        real_idx = i.pop(min_vert)

        # cntr = 3
        while this_area < np.inf:
            '''min_vert was removed from areas and i.  Now,
            adjust the adjacent areas and remove the new
            min_vert.
            Now that min_vert was filtered out, min_vert points
            to the point after the deleted point.'''

            skip = False  # modified area may be the next minvert

            try:
                right_area = triangle_area(pts[i[min_vert - 1]],
                                           pts[i[min_vert]], pts[i[min_vert + 1]])
            except IndexError:
                # trying to update area of endpoint. Don't do it
                pass
            else:
                right_idx = i[min_vert]
                if right_area <= this_area:
                    # even if the point now has a smaller area,
                    # it ultimately is not more significant than
                    # the last point, which needs to be removed
                    # first to justify removing this point.
                    # Though this point is the next most significant
                    right_area = this_area

                    # min_vert refers to the point to the right of
                    # the previous min_vert, so we can leave it
                    # unchanged if it is still the min_vert
                    skip = min_vert

                # update both collections of areas
                real_areas[right_idx] = right_area
                areas[min_vert] = right_area

            if min_vert > 1:
                # cant try/except because 0-1=-1 is a valid index
                left_area = triangle_area(pts[i[min_vert - 2]],
                                          pts[i[min_vert - 1]], pts[i[min_vert]])
                if left_area <= this_area:
                    # same justification as above
                    left_area = this_area
                    skip = min_vert - 1
                real_areas[i[min_vert - 1]] = left_area
                areas[min_vert - 1] = left_area

            # only argmin if we have too.
            min_vert = skip or argmin(areas)
            real_idx = i.pop(min_vert)
            this_area = areas[min_vert]
            # areas = np.delete(areas,min_vert) #slower
            remove(areas, min_vert)  # faster
            '''if sum(np.where(areas==np.inf)[0]) != sum(list(reversed(range(len(areas))))[:cntr]):
              print "broke:",np.where(areas==np.inf)[0],cntr
              break
            cntr+=1
            #if real_areas[0]<np.inf or real_areas[-1]<np.inf:
            #  print "NO!", real_areas[0], real_areas[-1]
            '''
        return real_areas

    def from_threshold(self, threshold):
        return self.pts[self.thresholds >= threshold]

    def from_number(self, n):
        thresholds = self.ordered_thresholds
        try:
            threshold = thresholds[int(n)]
        except IndexError:
            return self.pts
        return self.pts[self.thresholds > threshold]

    def from_ratio(self, r):
        if r <= 0 or r > 1:
            raise ValueError("Ratio must be 0<r<=1")
        else:
            return self.from_number(r * len(self.thresholds))


class WKTSimplifier(VWSimplifier):
    """
    VWSimplifier that returns strings suitable for WKT creation
    """

    def __init__(self, *args, **kwargs):
        if 'precision' in kwargs:
            p = kwargs.pop('precision')
        else:
            p = None
        VWSimplifier.__init__(self, *args, **kwargs)
        self.set_precision(p)

    def set_precision(self, precision):
        if precision:
            self.pts_as_strs = self.pts.astype('S%s' % precision)
        else:
            self.pts_as_strs = self.pts.astype(str)

    '''slow
    def from_threshold(self,threshold,precision=None):
        arr = np.array2string(self.pts[self.thresholds > threshold],precision=precision)
        return arr.replace('[[ ','(').replace(']]',')').replace(']\n [ ',',')
    '''

    def wkt_from_threshold(self, threshold, precision=None):
        if precision:
            self.set_precision(precision)
        pts = self.pts_as_strs[self.thresholds >= threshold]
        return '(%s)' % ','.join(['%s %s' % (x, y) for x, y in pts])

    def wkt_from_number(self, n, precision=None):
        thresholds = self.ordered_thresholds
        if n < 3: n = 3  # For polygons. TODO something better
        try:
            threshold = thresholds[int(n)]
        except IndexError:
            threshold = 0

        return self.wkt_from_threshold(threshold, precision=precision)

    def wkt_from_ratio(self, r, precision=None):
        if r <= 0 or r > 1:
            raise ValueError("Ratio must be 0<r<=1")
        else:
            return self.wkt_from_number(r * len(self.thresholds))


try:
    from django.contrib.gis.gdal import OGRGeometry, OGRException
    from django.contrib.gis.geos import GEOSGeometry, fromstr
except ImportError:
    class GDALSimplifier(object):
        """
        Dummy object that would be replaced by a real one if correct module exists
        """

        def __init__(*args, **kwargs):
            log.info("""
                  django.contrib.gis.gdal not found.
                  GDALSimplifier not available.
                  """)
else:
    from json import loads
    import re

    p = re.compile('([ 0123456789.]+) ([0123456789.]+)')


    class GDALSimplifier(object):
        """
        Warning, there is a slight loss of precision just in the
        conversion from geometry object to numpy.array even if no
        threshold is applied.  ie:

        originalpolygeom.area   ->   413962.65495176613
        gdalsimplifierpoly.area ->   413962.65495339036
        """

        def __init__(self, geom, precision=None, return_GDAL=True):
            """
            accepts a gdal.OGRGeometry or geos.GEOSGeometry
            object and wraps multiple
            VWSimplifiers.  set return_GDAL to False for faster
            filtering with arrays of floats returned instead of
            geometry objects.
            """
            global p
            self.return_GDAL = return_GDAL
            if isinstance(geom, OGRGeometry):
                name = geom.geom_name
                self.Geometry = lambda w: OGRGeometry(w, srs=geom.srs)
                self.pts = np.array(geom.tuple)
            elif isinstance(geom, GEOSGeometry):
                name = geom.geom_type.upper()
                self.Geometry = lambda w: fromstr(w)
                self.pts = np.array(geom.tuple)
            elif isinstance(geom, str) or isinstance(geom, bytes):
                # assume wkt
                # for WKT
                def str2tuple(q):
                    return '(%s,%s)' % (q.group(1), q.group(2))

                self.return_GDAL = False  # don't even try
                self.Geometry = lambda w: w  # this will never be used
                name, pts = geom.split(' ', 1)
                self.pts = loads(p.sub(str2tuple, pts). \
                                 replace('(', '[').replace(')', ']'))
            self.precision = precision
            if name == 'LINESTRING':
                self.maskfunc = self.linemask
                self.buildfunc = self.linebuild
                self.fromnumfunc = self.notimplemented
            elif name == "POLYGON":
                self.maskfunc = self.polymask
                self.buildfunc = self.polybuild
                self.fromnumfunc = self.notimplemented
            elif name == "MULTIPOLYGON":
                self.maskfunc = self.multimask
                self.buildfunc = self.multibuild
                self.fromnumfunc = self.notimplemented
            else:
                raise RuntimeError("""
             Only types LINESTRING, POLYGON and MULTIPOLYGON
             supported, but got %s""" % name)
            # sets self.simplifiers to a list of VWSimplifiers
            self.buildfunc()

        # rather than concise, I'd rather be explicit and clear.

        def pt2str(self, pt):
            """make length 2 numpy.array.__str__() fit for wkt"""
            return ' '.join(pt)

        def linebuild(self):
            self.simplifiers = [WKTSimplifier(self.pts)]

        def line2wkt(self, pts):
            return u'LINESTRING %s' % pts

        def linemask(self, threshold):
            get_pts = self.get_pts
            pts = get_pts(self.simplifiers[0], threshold)
            if self.return_GDAL:
                return self.Geometry(self.line2wkt(pts))
            else:
                return pts

        def polybuild(self):
            list_of_pts = self.pts
            result = []
            for pts in list_of_pts:
                result.append(WKTSimplifier(pts))
            self.simplifiers = result

        def poly2wkt(self, list_of_pts):
            return u'POLYGON (%s)' % ','.join(list_of_pts)

        def polymask(self, threshold):
            get_pts = self.get_pts
            sims = self.simplifiers
            list_of_pts = [get_pts(sim, threshold) for sim in sims]
            if self.return_GDAL:
                return self.Geometry(self.poly2wkt(list_of_pts))
            else:
                return array(list_of_pts)

        def multibuild(self):
            list_of_list_of_pts = self.pts
            result = []
            for list_of_pts in list_of_list_of_pts:
                subresult = []
                for pts in list_of_pts:
                    subresult.append(WKTSimplifier(pts))
                result.append(subresult)
            self.simplifiers = result

        def multi2wkt(self, list_of_list_of_pts):
            outerlist = []
            for list_of_pts in list_of_list_of_pts:
                outerlist.append('(%s)' % ','.join(list_of_pts))
            return u'MULTIPOLYGON (%s)' % ','.join(outerlist)

        def multimask(self, threshold):
            loflofsims = self.simplifiers
            result = []
            get_pts = self.get_pts
            if self.return_GDAL:
                ret_func = lambda r: self.Geometry(self.multi2wkt(r))
            else:
                ret_func = lambda r: r
            for list_of_simplifiers in loflofsims:
                subresult = []
                for simplifier in list_of_simplifiers:
                    subresult.append(get_pts(simplifier, threshold))
                result.append(subresult)
            return ret_func(result)

        def notimplemented(self, n):
            log.info('This function is not yet implemented')

        def from_threshold(self, threshold):
            precision = self.precision
            if self.return_GDAL:
                self.get_pts = lambda obj, t: obj.wkt_from_threshold(t, precision)
            else:
                self.get_pts = lambda obj, t: obj.from_threshold(t)
            return self.maskfunc(threshold)

        def from_number(self, n):
            precision = self.precision
            if self.return_GDAL:
                self.get_pts = lambda obj, t: obj.wkt_from_number(t, precision)
            else:
                self.get_pts = lambda obj, t: obj.from_number(t)
            return self.maskfunc(n)

        def from_ratio(self, r):
            precision = self.precision
            if self.return_GDAL:
                self.get_pts = lambda obj, t: obj.wkt_from_ratio(t, precision)
            else:
                self.get_pts = lambda obj, t: obj.from_ratio(t)
            return self.maskfunc(r)


def fancy_parametric(k):
    """
    good k's: .33,.5,.65,.7,1.3,1.4,1.9,3,4,5
    """
    cos = np.cos
    sin = np.sin
    xt = lambda t: (k - 1) * cos(t) + cos(t * (k - 1))
    yt = lambda t: (k - 1) * sin(t) - sin(t * (k - 1))
    return xt, yt


if __name__ == "__main__":
    from time import time

    n = 5000
    thetas = np.linspace(0, 16 * np.pi, n)
    xt, yt = fancy_parametric(1.4)
    pts = np.array([[xt(t), yt(t)] for t in thetas])
    start = time()
    simplifier = VWSimplifier(pts)
    pts = simplifier.from_number(1000)
    end = time()
    log.info("%s vertices removed in %02f seconds" % (n - len(pts), end - start))

    import matplotlib

    matplotlib.use('AGG')
    import matplotlib.pyplot as plot

    plot.plot(pts[:, 0], pts[:, 1], color='r')
    plot.savefig('visvalingam.png')
    log.info("saved visvalingam.png")
    # plot.show()



###############################################################################
# general GAMMA utilities

# Copyright (c) 2014-2025, the pyroSAR Developers, Stefan Engelhardt.

# This file is part of the pyroSAR Project. It is subject to the
# license terms in the LICENSE.txt file found in the top-level
# directory of this distribution and at
# https://github.com/johntruckenbrodt/pyroSAR/blob/master/LICENSE.txt.
# No part of the pyroSAR project, including this file, may be
# copied, modified, propagated, or distributed except according
# to the terms contained in the LICENSE.txt file.
################################################################################
import math
import os
import re
import string
import codecs
import subprocess as sp
from datetime import datetime

from pyroSAR.examine import ExamineGamma
from spatialist.ancillary import parse_literal, run, union, dissolve
from spatialist.envi import hdr

from .error import gammaErrorHandler


def do_execute(par, ids, exist_ok):
    """
    small helper function to assess whether a GAMMA command shall be executed.

    Parameters
    ----------
    par: dict
        a dictionary containing all arguments for the command
    ids: list[str]
        the IDs of the output files
    exist_ok: bool
        allow existing output files?

    Returns
    -------
    bool
        execute the command because (a) not all output files exist or (b) existing files are not allowed
    """
    all_exist = all([os.path.isfile(par[x]) for x in ids if par[x] != '-'])
    return (exist_ok and not all_exist) or not exist_ok


class ISPPar(object):
    """
    Reader for ISP parameter files of the GAMMA software package

    This class allows to read all information from files in GAMMA's parameter file format.
    Each key-value pair is parsed and added as attribute. For instance if the parameter file
    contains the pair 'sensor:    TSX-1' an attribute named 'sensor' with the value 'TSX-1' will be available.

    The values are converted to native Python types, while unit identifiers like 'dB' or 'Hz' are removed.
    Please see the GAMMA reference manual for further information on the actual file format.
    
    Parameters
    ----------
    filename: str
        the GAMMA parameter file
    
    Examples
    --------
    >>> from pyroSAR.gamma import ISPPar
    >>> with ISPPar('S1A__IW___A_20141115T181801_VH_grd.par') as par:
    ...     print(par) # print an overview of all available metadata
    ...     print(par.keys) # print all parameter names
    ...     for key, value in par.envidict().items():
    ...         print('{0}: {1}'.format(key, value)) # print the ENVI HDR compliant metadata
    
    Attributes
    ----------
    keys: list
        the names of all parameters
    """
    
    _re_kv_pair = re.compile(r'^(\w+):\s*(.+)\s*')
    _re_float_literal = re.compile(r'^[+-]?(?:(\d*\.\d+)|(\d+\.?))(?:[Ee][+-]?\d+)?')
    
    def __init__(self, filename):
        """Parses an ISP parameter file from disk.

        Args:
            filename: The filename or file object representing the ISP parameter file.
        """
        if isinstance(filename, str):
            par_file = open(filename, 'r')
        else:
            par_file = filename
        
        self.keys = ['filetype']
        
        try:
            content = par_file.read().split('\n')
        except UnicodeDecodeError:
            par_file = codecs.open(filename, 'r', encoding='utf-8', errors='ignore')
            content = par_file.read()
            printable = set(string.printable)
            content = filter(lambda x: x in printable, content)
            content = ''.join(list(content)).split('\n')
        finally:
            par_file.close()
        
        if 'Image Parameter File' in content[0]:
            setattr(self, 'filetype', 'isp')
        elif 'DEM/MAP parameter file' in content[0]:
            setattr(self, 'filetype', 'dem')
        else:
            raise RuntimeError('unknown parameter file type')
        
        for line in content:
            match = ISPPar._re_kv_pair.match(line)
            if not match:
                continue  # Skip malformed lines with no key-value pairs
            key = match.group(1)
            items = match.group(2).split()
            if len(items) == 0:
                value = None
            elif len(items) == 1:
                value = parse_literal(items[0])
            else:
                if not ISPPar._re_float_literal.match(items[0]):
                    # Value is a string literal containing whitespace characters
                    value = match.group(2)
                else:
                    # Evaluate each item and stop at the first non-float literal
                    value = []
                    for i in items:
                        match = ISPPar._re_float_literal.match(i)
                        if match:
                            value.append(parse_literal(match.group()))
                        else:
                            # If the first float literal is immediately followed by a non-float literal handle the
                            # first one as singular value, e.g. in '20.0970 dB'
                            if len(value) == 1:
                                value = value[0]
                            break
            self.keys.append(key)
            setattr(self, key, value)
        
        if hasattr(self, 'date'):
            try:
                self.date = '{}-{:02d}-{:02d}T{:02d}:{:02d}:{:02f}'.format(*self.date)
            except:
                # if only date available
                self.date = '{}-{:02d}-{:02d}'.format(*self.date)
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        return
    
    def __getattr__(self, item):
        # will only be run if object has no attribute item
        raise AttributeError("parameter file has no attribute '{}'".format(item))
    
    def __str__(self):
        maxlen = len(max(self.keys, key=len)) + 1
        return '\n'.join(['{key}:{sep}{value}'.format(key=key,
                                                      sep=(maxlen - len(key)) * ' ',
                                                      value=getattr(self, key)) for key in self.keys])
    
    def envidict(self, nodata=None):
        """
        export relevant metadata to an ENVI HDR file compliant format
        
        Parameters
        ----------
        nodata: int, float or None
            a no data value to write to the HDR file via attribute 'data ignore value'
        
        Returns
        -------
        dict
            a dictionary containing attributes translated to ENVI HDR naming
        """
        out = dict(bands=1,
                   header_offset=0,
                   file_type='ENVI Standard',
                   interleave='bsq',
                   sensor_type='Unknown',
                   byte_order=1,
                   wavelength_units='Unknown')
        
        if hasattr(self, 'date'):
            out['acquisition_time'] = self.date + 'Z'
        
        out['samples'] = getattr(self, union(['width', 'range_samples', 'samples'], self.keys)[0])
        out['lines'] = getattr(self, union(['nlines', 'azimuth_lines', 'lines'], self.keys)[0])
        
        dtypes_lookup = {'FCOMPLEX': 6, 'FLOAT': 4, 'REAL*4': 4, 'INTEGER*2': 2, 'SHORT': 12}
        dtype = getattr(self, union(['data_format', 'image_format'], self.keys)[0])
        
        if dtype not in dtypes_lookup.keys():
            raise TypeError('unsupported data type: {}'.format(dtype))
        
        out['data_type'] = dtypes_lookup[dtype]
        
        if nodata is not None:
            out['data_ignore_value'] = nodata
        
        if out['data_type'] == 6:
            out['complex_function'] = 'Power'
        # projections = ['AEAC', 'EQA', 'LCC', 'LCC2', 'OMCH', 'PC', 'PS', 'SCH', 'TM', 'UTM']
        # the corner coordinates are shifted by 1/2 pixel to the Northwest since GAMMA pixel
        # coordinates are defined for the pixel center while in ENVI it is the upper left
        if hasattr(self, 'DEM_projection'):
            if self.DEM_projection == 'UTM':
                hem = 'North' if float(self.false_northing) == 0 else 'South'
                out['map_info'] = ['UTM', '1.0000', '1.0000',
                                   self.corner_east - (abs(self.post_east) / 2),
                                   self.corner_north + (abs(self.post_north) / 2),
                                   str(abs(float(self.post_east))),
                                   str(abs(float(self.post_north))),
                                   self.projection_zone, hem, 'WGS-84', 'units=Meters']
            elif self.DEM_projection == 'EQA':
                out['map_info'] = ['Geographic Lat/Lon', '1.0000', '1.0000',
                                   self.corner_lon - (abs(self.post_lon) / 2),
                                   self.corner_lat + (abs(self.post_lat) / 2),
                                   str(abs(float(self.post_lon))),
                                   str(abs(float(self.post_lat))),
                                   'WGS-84', 'units=Degrees']
            elif self.DEM_projection == 'PS':
                if self.projection_name == 'WGS 84 / Antarctic Polar Stereographic':
                    out['map_info'] = [
                        'EPSG:3031 - WGS 84 / Antarctic Polar Stereographic',
                        '1.0000',
                        '1.0000',
                        self.corner_east - (abs(self.post_east) / 2),
                        self.corner_north + (abs(self.post_north) / 2),
                        str(abs(float(self.post_east))),
                        str(abs(float(self.post_north))),
                        'WGS-84',
                        'units=Meters',
                    ]
                elif self.projection_name == 'WGS 84 / Arctic Polar Stereographic':
                    out['map_info'] = [
                        'EPSG:3995 - WGS 84 / Arctic Polar Stereographic',
                        '1.0000',
                        '1.0000',
                        self.corner_east - (abs(self.post_east) / 2),
                        self.corner_north + (abs(self.post_north) / 2),
                        str(abs(float(self.post_east))),
                        str(abs(float(self.post_north))),
                        'WGS-84',
                        'units=Meters',
                    ]
                else:
                    raise RuntimeError(
                        f'unsupported projection: "{self.DEM_projection}; {self.projection_name}". The projection name "{self.projection_name}" was not recognised. Expected projection names are "WGS 84 / Arctic Polar Stereographic" and "WGS 84 / Antarctic Polar Stereographic". Add support for the required projection name as an ENVI map info output in gamma.auxil.ISPPar.envidict.'
                    )
            else:
                raise RuntimeError(
                    f'unsupported projection: "{self.DEM_projection}; {self.projection_name}". To resolve, create an ENVI map info output for this projection in gamma.auxil.ISPPar.envidict.'
                )
        return out


class Namespace(object):
    """
    GAMMA file name handler. This improves managing the many files names
    handled when processing with GAMMA.
    
    Parameters
    ----------
    directory: str
        the directory path where files shall be written.
    basename: str
        the product basename as returned by
        :meth:`pyroSAR.drivers.ID.outname_base`
    
    Examples
    --------
    >>> n = Namespace(directory='/path', basename='S1A__IW___A_20180829T170631')
    >>> print(n.pix_geo)
    '-'
    >>> n.appreciate(['pix_geo'])
    >>> print(n.pix_geo)
    '/path/S1A__IW___A_20180829T170631_pix_geo'
    """
    def __init__(self, directory, basename):
        self.__base = basename
        self.__outdir = directory
        self.__reg = []
    
    def __getitem__(self, item):
        item = str(item).replace('.', '_')
        return self.get(item)
    
    def __getattr__(self, item):
        # will only be run if object has no attribute item
        return '-'
    
    def appreciate(self, keys):
        """

        Parameters
        ----------
        keys: list[str]

        Returns
        -------

        """
        for key in keys:
            setattr(self, key.replace('.', '_'), os.path.join(self.__outdir, self.__base + '_' + key))
            if key not in self.__reg:
                self.__reg.append(key.replace('.', '_'))
    
    def depreciate(self, keys):
        """

        Parameters
        ----------
        keys: list[str]

        Returns
        -------

        """
        for key in keys:
            setattr(self, key.replace('.', '_'), '-')
            if key not in self.__reg:
                self.__reg.append(key.replace('.', '_'))
    
    def getall(self):
        out = {}
        for key in self.__reg:
            out[key] = getattr(self, key)
        return out
    
    def select(self, selection):
        return [getattr(self, key) for key in selection]
    
    def isregistered(self, key):
        return key in self.__reg
    
    def isappreciated(self, key):
        if self.isregistered(key):
            if self.get(key) != '-':
                return True
        return False
    
    def isfile(self, key):
        return hasattr(self, key) and os.path.isfile(getattr(self, key))
    
    def get(self, key):
        return getattr(self, key)


def par2hdr(parfile, hdrfile, modifications=None, nodata=None):
    """
    Create an ENVI HDR file from a GAMMA PAR file
    
    Parameters
    ----------
    parfile: str
        the GAMMA parfile
    hdrfile: str
        the ENVI HDR file
    modifications: dict or None
        a dictionary containing value deviations to write to the HDR file
    nodata: int, float or None
        a no data value to write to the HDR file via attribute 'data ignore value'

    Returns
    -------
    
    Examples
    --------
    >>> from pyroSAR.gamma.auxil import par2hdr
    >>> par2hdr('dem_seg.par', 'inc.hdr')
    # write a HDR file for byte data based on a parfile of float data
    >>> par2hdr('dem_seg.par', 'ls_map.hdr', modifications={'data_type': 1})
    
    See Also
    --------
    :class:`spatialist.envi.HDRobject`
    :func:`spatialist.envi.hdr`
    """
    
    with ISPPar(parfile) as par:
        items = par.envidict(nodata)
        if modifications is not None:
            items.update(modifications)
        hdr(items, hdrfile)


def process(cmd, outdir=None, logfile=None, logpath=None,
            inlist=None, void=True, shellscript=None):
    """
    wrapper function to execute GAMMA commands via module :mod:`subprocess`

    Parameters
    ----------
    cmd: list[str]
        the command line arguments
    outdir: str
        the directory to execute the command in
    logfile: str
        a file to write the command log to; overrides parameter logpath
    logpath: str
        a directory to write logfiles to; the file will be named {GAMMA command}.log, e.g. gc_map.log;
        is overridden by parameter logfile
    inlist: list
        a list of values, which is passed as interactive inputs via stdin
    void: bool
        return the stdout and stderr messages?
    shellscript: str
        a file to write the GAMMA commands to in shell format

    Returns
    -------
    tuple of str or None
        the stdout and stderr messages if void is False, otherwise None
    """
    if logfile is not None:
        log = logfile
    else:
        log = os.path.join(logpath, os.path.basename(cmd[0]) + '.log') if logpath else None
    gamma_home = ExamineGamma().home
    if shellscript is not None:
        if not os.path.isfile(shellscript):
            # create an empty file
            with open(shellscript, 'w') as init:
                pass
        line = ' '.join([str(x) for x in dissolve(cmd)])
        if inlist is not None:
            line += ' <<< $"{}"'.format('\n'.join([str(x) for x in inlist]) + '\n')
        with open(shellscript, 'r+') as sh:
            if outdir is not None:
                content = sh.read()
                sh.seek(0)
                is_new = re.search('this script was created automatically by pyroSAR', content) is None
                if is_new:
                    ts = datetime.now().strftime('%a %b %d %H:%M:%S %Y')
                    sh.write('# this script was created automatically by pyroSAR on {}\n\n'.format(ts))
                    sh.write('export base={}\n'.format(outdir))
                    sh.write('export GAMMA_HOME={}\n\n'.format(gamma_home))
                    sh.write(content)
                line = line.replace(outdir, '$base').replace(gamma_home, '$GAMMA_HOME')
            sh.seek(0, 2)  # set pointer to the end of the file
            sh.write(line + '\n\n')
    
    # create an environment containing the locations of all GAMMA submodules to be passed to the subprocess calls
    gammaenv = os.environ.copy()
    gammaenv['GAMMA_HOME'] = gamma_home
    out, err = run([ExamineGamma().gdal_config, '--datadir'], void=False)
    gammaenv['GDAL_DATA'] = out.strip()
    for module in ['DIFF', 'DISP', 'IPTA', 'ISP', 'LAT']:
        loc = os.path.join(gammaenv['GAMMA_HOME'], module)
        if os.path.isdir(loc):
            gammaenv[module + '_HOME'] = loc
            for submodule in ['bin', 'scripts']:
                subloc = os.path.join(loc, submodule)
                if os.path.isdir(subloc):
                    gammaenv['PATH'] += os.pathsep + subloc
    
    # execute the command
    out, err = run(cmd, outdir=outdir, logfile=log, inlist=inlist, void=False, errorpass=True, env=gammaenv)
    gammaErrorHandler(out, err)
    if not void:
        return out, err


def slc_corners(parfile):
    """
    extract the corner coordinates of a SAR scene

    Parameters
    ----------
    parfile: str
        the GAMMA parameter file to read coordinates from

    Returns
    -------
    dict of float
        a dictionary with keys xmin, xmax, ymin, ymax
    """
    out, err = process(['SLC_corners', parfile], void=False)
    pts = {}
    pattern = r'-?[0-9]+\.[0-9]+'
    for line in out.split('\n'):
        if line.startswith('min. latitude'):
            pts['ymin'], pts['ymax'] = [float(x) for x in
                                        re.findall(pattern, line)]
        elif line.startswith('min. longitude'):
            pts['xmin'], pts['xmax'] = [float(x) for x in
                                        re.findall(pattern, line)]
    return pts


class Spacing(object):
    """
    compute multilooking factors and pixel spacings from an ISPPar object for a defined ground range target pixel spacing

    Parameters
    ----------
    par: str or ISPPar
        the ISP parameter file
    spacing: int or float
        the target pixel spacing in ground range
    """
    
    def __init__(self, par, spacing='automatic'):
        # compute ground range pixel spacing
        par = par if isinstance(par, ISPPar) else ISPPar(par)
        self.groundRangePS = par.range_pixel_spacing / (math.sin(math.radians(par.incidence_angle)))
        # compute initial multilooking factors
        if spacing == 'automatic':
            if self.groundRangePS > par.azimuth_pixel_spacing:
                ratio = self.groundRangePS / par.azimuth_pixel_spacing
                self.rlks = 1
                self.azlks = int(round(ratio))
            else:
                ratio = par.azimuth_pixel_spacing / self.groundRangePS
                self.rlks = int(round(ratio))
                self.azlks = 1
        else:
            self.rlks = int(round(float(spacing) / self.groundRangePS))
            self.azlks = int(round(float(spacing) / par.azimuth_pixel_spacing))


class UTM(object):
    """
    convert a gamma parameter file corner coordinate from EQA to UTM
    
    Parameters
    ----------
    parfile: str
        the GAMMA parameter file to read the coordinate from
    
    Example
    -------
    
    >>> from pyroSAR.gamma import UTM
    >>> print(UTM('gamma.par').zone)
    """
    
    def __init__(self, parfile):
        par = ISPPar(parfile)
        inlist = ['WGS84', 1, 'EQA', par.corner_lon, par.corner_lat, '', 'WGS84', 1, 'UTM', '']
        inlist = map(str, inlist)
        proc = sp.Popen(['coord_trans'], stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE,
                        universal_newlines=True, shell=False)
        out, err = proc.communicate(''.join([x + '\n' for x in inlist]))
        out = [x for x in filter(None, out.split('\n')) if ':' in x]
        
        self.meta = dict()
        for line in out:
            key, value = re.split(r'\s*:\s*', line)
            value = value.split()
            value = map(parse_literal, value) if len(value) > 1 else value[0]
            self.meta[key] = value
        try:
            self.zone, self.northing, self.easting, self.altitude = \
                self.meta['UTM zone/northing/easting/altitude (m)']
        except KeyError:
            self.zone, self.northing, self.easting = \
                self.meta['UTM zone/northing/easting (m)']
